import boto3

from dynamo_pandas.serde import TypeDeserializer
from dynamo_pandas.serde import TypeSerializer

ts = TypeSerializer()
td = TypeDeserializer()


def _deserialize(items):
    """Convert dictionaries to DynamoDB format and back."""
    return td.deserialize(ts.serialize(items))


def _batches(items, batch_size):
    """Split an iterable in batches."""
    items = list(items)
    start = 0
    while start < len(items):
        end = min(start + batch_size, len(items))
        yield items[start:end]
        start += batch_size


def get_item(*, key, table, attributes=None):
    """Get a single item from a table.

    Parameters
    ----------
    key : dict
        Key of the item to get.

    table : str
        Name of the DynamoDB table.

    attributes : list[str]
        Names of the item attributes to return. If None (default), all attributes are
        returned.

    Returns
    -------
    dict, None
        A dictionary representing the item's attributes. None if the key does not exist
        in the table.

    Examples
    --------

    >>> item = get_item(key={"player_id": "player_two"}, table="players")
    >>> print(item)
    {'bonus_points': 1,
     'player_id': 'player_two',
     'last_play': '2021-01-19 19:07:54',
     'rating': 3.8,
     'play_time': '0 days 22:07:34'}

    Get only specific attributes:

    >>> item = get_item(
    ...     key={"player_id": "player_two"},
    ...     table="players",
    ...     attributes=["play_time", "rating"]
    ... )
    >>> print(item)
    {'rating': 3.8, 'play_time': '0 days 22:07:34'}
    """
    table = boto3.resource("dynamodb").Table(table)

    kwargs = {}
    if attributes is not None:
        kwargs["ProjectionExpression"] = ", ".join(attributes)

    item = table.get_item(Key=key, **kwargs).get("Item")

    return _deserialize(item)


def get_items(*, keys, table, attributes=None):
    """Get multiple items from a table.

    Parameters
    ----------
    keys : list[dict]
        List of key dictionaries of the items to get.

    table : str
        Name of the DynamoDB table.

    attributes : list[str]
        Names of the item attributes to return. If None (default), all attributes are
        returned.

    Returns
    -------
    list[dict]
        List of dictionaties where each dictionary represents an item's attributes.
        Only items for which the key exists in the table are returned.

    Examples
    --------

    >>> items = get_items(
    ...     keys=[
    ...         {"player_id": "player_two"},
    ...         {"player_id": "player_one"},
    ...         {"player_id": "player_five"}, # Not in the table
    ...     ],
    ...     table="players"
    ... )
    >>> print(items)
    [{'bonus_points': 3, 'player_id': 'player_one', 'last_play': '2021-01-18 22:47:23', 'rating': 4.3, 'play_time': '2 days 17:41:55'},
     {'bonus_points': 1, 'player_id': 'player_two', 'last_play': '2021-01-19 19:07:54', 'rating': 3.8, 'play_time': '0 days 22:07:34'}]

    Get only specific attributes:

    >>> items = get_items(
    ...     keys=[{"player_id": "player_two"}, {"player_id": "player_one"}],
    ...     table="players",
    ...     attributes=["player_id", "play_time"]
    ... )
    >>> print(items)
    [{'player_id': 'player_one', 'play_time': '2 days 17:41:55'}, {'player_id': 'player_two', 'play_time': '0 days 22:07:34'}]
    """  # noqa: E501

    def _request(keys, table=table, attributes=attributes):
        table_dict = {"Keys": keys}

        if attributes is not None:
            table_dict["ProjectionExpression"] = ", ".join(attributes)

        return {table: table_dict}

    def _get_items(keys, table=table, attributes=attributes):
        response = resource.batch_get_item(RequestItems=_request(keys))
        items = response["Responses"][table]

        while response["UnprocessedKeys"] != {}:
            keys = response["UnprocessedKeys"][table]["Keys"]
            response = resource.batch_get_item(RequestItems=_request(keys))
            items.extend(response["Responses"][table])

        return items

    resource = boto3.resource("dynamodb")

    key_batches = _batches(keys, batch_size=100)

    items = []
    for key_batch in key_batches:
        items.extend(_get_items(key_batch))

    return _deserialize(items)


def get_all_items(*, table, attributes=None):
    """Get all the items in a table.

    This function performs a scan of the table.

    Parameters
    ----------
    table : str
        Name of the DynamoDB table.

    attributes : list[str]
        Names of the item attributes to return. If None (default), all attributes are
        returned.

    Returns
    -------
    list[dict]
        List of dictionaties where each dictionary represents an item's attributes.

    Examples
    --------

    >>> items = get_all_items(table="players")
    >>> print(items)
    [{'bonus_points': 4, 'player_id': 'player_three', 'last_play': '2021-01-21 10:22:43', 'rating': 2.5, 'play_time': '1 days 14:01:19'},
     {'bonus_points': None, 'player_id': 'player_four', 'last_play': '2021-01-22 13:51:12', 'rating': 4.8, 'play_time': '0 days 03:45:49'},
     {'bonus_points': 3, 'player_id': 'player_one', 'last_play': '2021-01-18 22:47:23', 'rating': 4.3, 'play_time': '2 days 17:41:55'},
     {'bonus_points': 1, 'player_id': 'player_two', 'last_play': '2021-01-19 19:07:54', 'rating': 3.8, 'play_time': '0 days 22:07:34'}]

    Get only specific attributes:

    >>> items = get_all_items(table="players", attributes=["player_id", "play_time"])
    >>> print(items)
    [{'player_id': 'player_three', 'play_time': '1 days 14:01:19'},
     {'player_id': 'player_four', 'play_time': '0 days 03:45:49'},
     {'player_id': 'player_one', 'play_time': '2 days 17:41:55'},
     {'player_id': 'player_two', 'play_time': '0 days 22:07:34'}]
    """  # noqa: E501
    table = boto3.resource("dynamodb").Table(table)

    kwargs = {}
    if attributes is not None:
        kwargs["ProjectionExpression"] = ", ".join(attributes)

    response = table.scan(**kwargs)

    items = response["Items"]

    while "LastEvaluatedKey" in response:
        response = table.scan(ExclusiveStartKey=response["LastEvaluatedKey"], **kwargs)
        items.extend(response["Items"])

    return _deserialize(items)


def put_item(*, item, table, return_response=False):
    """Add or update an item in a table. If the item does not exist in the table it is
    created, otherwise the existing item is replaced with the new one.

    Item can use supported numpy or pandas data types.

    Parameters
    ----------
    item : dict
        A dictionary representing the item's attributes. The item must include the
        table's primary key attributes.

    table : str
        Name of the DynamoDB table.

    return_response : bool
        If True, the response from the boto3 API call will be returned.

    Returns
    -------
    None, dict
        None if ``return_response`` is False, the boto3 API call response if True.

    Examples
    --------

    >>> print(item)
    {'player_id': 'player_three',
     'bonus_points': 4,
     'last_play': Timestamp('2021-01-21 10:22:43+0000', tz='UTC'),
     'rating': 2.5,
     'play_time': Timedelta('1 days 14:01:19')}
    >>> response = put_item(item=item, table="players", return_response=True)
    >>> print(response["ResponseMetadata"]["HTTPStatusCode"])
    200
    """
    if not isinstance(item, dict):
        raise TypeError("item must be a non-empty dictionary")

    client = boto3.client("dynamodb")

    response = client.put_item(TableName=table, Item=ts.serialize(item)["M"])

    if return_response:
        return response


def _put_items(items, table):
    """Adapter function to format the items, call the client batch_write_item function
    and return the unprocessed items (if any) in the format they were provided."""
    client = boto3.client("dynamodb")

    response = client.batch_write_item(
        RequestItems={table: [{"PutRequest": {"Item": item}} for item in items]}
    )

    if response["UnprocessedItems"] != {}:
        return [
            item["PutRequest"]["Item"] for item in response["UnprocessedItems"][table]
        ]
    else:
        return []


def put_items(*, items, table):
    """Add or update multiple items in a table. If the item(s) do not exist in the
    table they are created, otherwise the existing items are replaced with the new ones.

    Items can use supported numpy or pandas data types.

    Parameters
    ----------
    items : list[dict]
        List of dictionaries where each dictionary represents an item's attributes.

    table : str
        Name of the DynamoDB table.

    Examples
    --------

    >>> pprint(items)
    [{'bonus_points': 4,
      'last_play': Timestamp('2021-01-21 10:22:43+0000', tz='UTC'),
      'play_time': Timedelta('1 days 14:01:19'),
      'player_id': 'player_three',
      'rating': 2.5},
     {'bonus_points': <NA>,
      'last_play': Timestamp('2021-01-22 13:51:12+0000', tz='UTC'),
      'play_time': Timedelta('0 days 03:45:49'),
      'player_id': 'player_four',
      'rating': 4.8}]
    >>> put_items(items=items, table="players)
    """
    if not isinstance(items, list):
        raise TypeError("items must be a list of non-empty dictionaries")

    items_to_process = [i["M"] for i in ts.serialize(items)["L"]]

    batch_size = 25
    while len(items_to_process) > 0:
        batch_items = items_to_process[:batch_size]
        items_to_process = items_to_process[batch_size:]

        unprocessed_items = _put_items(batch_items, table)

        if len(unprocessed_items) > batch_size // 2:
            batch_size = max(batch_size // 2, 1)

        # Put unprocessed items at back of queue.
        items_to_process.extend(unprocessed_items)



from boto3.dynamodb.types import NULL
from boto3.dynamodb.types import NUMBER
from boto3.dynamodb.types import STRING
from boto3.dynamodb.types import TypeDeserializer as DDBTypeDeserializer
from boto3.dynamodb.types import TypeSerializer as DDBTypeSerializer
import numpy as np
import pandas as pd


class TypeSerializer(DDBTypeSerializer):
    """An extension of the `boto3.dynamodb.types.TypeSerializer
    <https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/dynamodb/types.html>`_
    class with support for pandas data types.

    Differences in type conversion from the parent class are the following::

        Python                                  DynamoDB
        ------                                  --------
        numpy.nan, pandas.NA                    {'NULL': True}
        int, numpy.integer                      {'N': str(value)}
        float, numpy.floating                   {'N': str(value)}
        pandas nullable Int64(32, 16, 8)        {'N': str(value)}
        pandas.Timestamp                        {'S': str(value)}
        pandas.Timedelta                        {'S': str(value)}
    """

    def _get_dynamodb_type(self, value):

        # Pandas NA values
        if value is pd.NA or value is pd.NaT:
            dynamodb_type = NULL

        elif isinstance(value, (pd.Timestamp, pd.Timedelta)):
            dynamodb_type = STRING

        elif isinstance(value, (int, np.integer)):
            dynamodb_type = NUMBER

        elif isinstance(value, (float, np.floating)):
            if np.isnan(value):
                dynamodb_type = NULL
            else:
                dynamodb_type = NUMBER

        else:
            dynamodb_type = super()._get_dynamodb_type(value)

        return dynamodb_type

    def _serialize_s(self, value):
        return str(value)

    def _serialize_n(self, value):
        if value % 1 == 0:
            v = int(value)
        else:
            v = str(value)
        return super()._serialize_n(v)


class TypeDeserializer(DDBTypeDeserializer):
    """An extension of the `boto3.dynamodb.types.TypeDeserializer
    <https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/dynamodb/types.html>`_
    class with conversion of numbers to float and int types instead of Decimal.

    Differences in type conversion from the parent class are the following::

        DynamoDB                                Python
        --------                                ------
        {'N': str(value)}                       int/float
    """

    def _deserialize_n(self, value):
        v = float(value)
        if v % 1 == 0:
            return int(v)
        else:
            return v



from .transactions import get_all_items
from .transactions import get_item
from .transactions import get_items
from .transactions import put_item
from .transactions import put_items

__all__ = ["get_all_items", "get_item", "get_items", "put_item", "put_items"]



import pandas as pd

from .transactions import get_all_items
from .transactions import get_items
from .transactions import put_items


def get_df(*, table, keys=None, attributes=None, dtype=None):
    """Get items from a table into a dataframe.

    Parameters
    ----------
    table : str
        Name of the DynamoDB table.

    keys : list[dict]
        List of keys to get where each key is represented by a dictionary.

    attributes : list[str]
        Names of the item attributes to return as dataframe columns. If None (default),
        all attributes are returned.

    dtype : data type or dict of column names -> data type
        Use a numpy.dtype or Python type to cast entire pandas object to the same type.
        Alternatively, use {col: dtype, …}, where col is a column label and dtype is a
        numpy.dtype or Python type to cast one or more of the DataFrame’s columns to
        column-specific types.

    Returns
    -------
    pandas.DataFrame
        A dataframe where each item from the table matching the requested keys is
        represented by a row and its attributes by columns.

    Examples
    --------

    >>> df = get_df(
    ...     table="players",
    ...     keys=[{"player_id": "player_three"}, {"player_id": "player_one"}]
    ... )
    >>> print(df)
       bonus_points     player_id            last_play  rating        play_time
    0             4  player_three  2021-01-21 10:22:43     2.5  1 days 14:01:19
    1             3    player_one  2021-01-18 22:47:23     4.3  2 days 17:41:55

    By default, the data types of the returned dataframe are basic pandas/numpy types:

    >>> df.info()
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 2 entries, 0 to 1
    Data columns (total 5 columns):
        #   Column        Non-Null Count  Dtype
       ---  ------        --------------  -----
        0   bonus_points  1 non-null      float64
        1   player_id     2 non-null      object
        2   last_play     2 non-null      object
        3   rating        2 non-null      float64
        4   play_time     2 non-null      object
    dtypes: float64(2), object(3)
    memory usage: 208.0+ bytes

    The ``dtype`` parameter can be used to specify the data types of the different
    columns:

    >>> df = get_df(
    ...     table="players",
    ...     keys=keys(player_id=["player_two", "player_four"]),
    ...         dtype={
    ...             "bonus_points": "Int8",
    ...             "last_play": "datetime64[ns, UTC]",
    ...             # "play_time": "timedelta64[ns]"  # See note below.
    ...         }
    ...     )
    >>> df.info()
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 2 entries, 0 to 1
    Data columns (total 5 columns):
        #   Column        Non-Null Count  Dtype
       ---  ------        --------------  -----
        0   bonus_points  1 non-null      Int8
        1   player_id     2 non-null      object
        2   last_play     2 non-null      datetime64[ns, UTC]
        3   rating        2 non-null      float64
        4   play_time     2 non-null      object
    dtypes: Int8(1), datetime64[ns, UTC](1), float64(1), object(2)
    memory usage: 196.0+ bytes

    .. note:: Due to a
        `known bug in pandas <https://github.com/pandas-dev/pandas/issues/38509>`_,
        timedelta strings cannot currently be converted back to timedelta64 type via the
        ``dtype`` parameter. Use the ``pandas.to_timedelta`` function instead:

        >>> df.play_time = pd.to_timedelta(df.play_time)
        >>> df.info()
        <class 'pandas.core.frame.DataFrame'>
        RangeIndex: 2 entries, 0 to 1
        Data columns (total 5 columns):
            #   Column        Non-Null Count  Dtype
           ---  ------        --------------  -----
            0   bonus_points  1 non-null      Int8
            1   player_id     2 non-null      object
            2   last_play     2 non-null      datetime64[ns, UTC]
            3   rating        2 non-null      float64
            4   play_time     2 non-null      timedelta64[ns]
        dtypes: Int8(1), datetime64[ns, UTC](1), float64(1), object(1), timedelta64[ns](1)
        memory usage: 196.0+ bytes

    Omitting the ``keys`` parameter performs a scan of the table and returns all the
    items.

    >>> df = get_df(table="players")
    >>> print(df)
        bonus_points     player_id            last_play  rating        play_time
    0           4.0  player_three  2021-01-21 10:22:43     2.5  1 days 14:01:19
    1           NaN   player_four  2021-01-22 13:51:12     4.8  0 days 03:45:49
    2           3.0    player_one  2021-01-18 22:47:23     4.3  2 days 17:41:55
    3           1.0    player_two  2021-01-19 19:07:54     3.8  0 days 22:07:34

    Specifying item attributes via the ``attributes`` parameter returns only the
    columns corresponding to the specified attributes:

    >>> df = get_df(table="players", attributes=["player_id", "rating"])
    >>> print(df)
          player_id  rating
    0  player_three     2.5
    1   player_four     4.8
    2    player_one     4.3
    3    player_two     3.8
    """  # noqa: E501
    if keys is not None:
        items = get_items(keys=keys, table=table, attributes=attributes)
    else:
        items = get_all_items(table=table, attributes=attributes)

    return _to_df(items=items, dtype=dtype)


def keys(**kwargs):
    """Generate a list of key dictionaries from the partition key attribute name and a
    list of values. This can simplify the generation of keys to use with the ``get_df``
    function when only a partition key is used.

    Parameters
    ----------
    **kwargs
        A single keyword argument corresponding to the partition key name with a value
        corresponding to the list of key values to return.

    Returns
    -------
    list[dict]
        A list of key dictionaries.

    Examples
    --------

    Assuming we have a table with ``player_id`` as the partition key, we can generate
    the list of keys from the list of players:

    >>> key_list = keys(player_id=["player_two", "player_three", "player_four"])
    >>> print(key_list)
    [{'player_id': 'player_one'}, {'player_id': 'player_three'}, {'player_id': 'player_four'}]
    """  # noqa: E501
    if len(kwargs.keys()) > 1:
        raise ValueError("Only one key attribute (partition key) is supported.")

    k = list(kwargs.keys())[0]
    return [{k: v} for v in kwargs[k]]


def put_df(df, *, table):
    """Put rows of a dataframe as items into a table. If the item(s) do not exist in the
    table they are created, otherwise the existing items are replaced with the new ones.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe of items to add/update in the table. The dataframe must, at a minimum,
        contain columns that correspond to the table's primary key attribute(s).

    table : str
        Name of the DynamoDB table.

    Examples
    --------
    Assume with have the following dataframe:

    >>> print(players_df)
          player_id           last_play       play_time  rating  bonus_points
    0    player_one 2021-01-18 22:47:23 2 days 17:41:55     4.3             3
    1    player_two 2021-01-19 19:07:54 0 days 22:07:34     3.8             1
    2  player_three 2021-01-21 10:22:43 1 days 14:01:19     2.5             4
    3   player_four 2021-01-22 13:51:12 0 days 03:45:49     4.8          <NA>

    The following will add or update the corresponding items in the table named
    ``players``:

    >>> put_df(players_df, table="players")
    """
    put_items(items=_to_items(df), table=table)


def _to_df(items, *, dtype=None):
    """Convert an item dictionary or list of item dictionaries into a pandas DataFrame.
    """
    if isinstance(items, dict):
        items = [items]

    df = pd.DataFrame(items)

    if dtype is not None:
        df = df.astype(dtype)

    return df


def _to_items(df):
    """Convert a pandas dataframe to a dictionary of items."""
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df must be a pandas DataFrame")

    return df.to_dict("records")



import codecs
from pathlib import Path
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    file_path = Path(__file__).parent / filename
    return codecs.open(file_path, encoding="utf-8").read()


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    re_str = r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta)
    meta_match = re.search(re_str, META_FILE, re.M)
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError(f"Unable to find __{meta}__ string.")


PACKAGE_NAME = "dynamo-pandas"
META_FILE = read(Path(__file__).parent / PACKAGE_NAME.replace("-", "_") / "__init__.py")


setup(
    name=PACKAGE_NAME,
    version=find_meta("version"),
    author="Julien de la Bruère-Terreault",
    author_email="drgfreeman@tuta.io",
    maintainer="Julien de la Bruère-Terreault",
    maintainer_email="drgfreeman@tuta.io",
    license="MIT",
    url="https://github.com/DrGFreeman/dynamo-pandas",
    description="Make working with pandas dataframe and AWS DynamoDB easy.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=["pandas>=1"],
    extras_require={"boto3": ["boto3"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
